@@ -16,7 +16,7 @@ import subprocess
 import sys
 import time
 import urllib
-from typing import Callable, Dict, Generator, List, Optional
+from typing import Callable, Dict, Generator, List, Literal, Optional, Union
 from uuid import uuid4
 
 
@@ -109,7 +109,7 @@ def await_summary(
     :param unchanged_timeout: The timeout in seconds, defaults to 60.
     :type unchanged_timeout: int, optional
     :return: The dictionary of the HTTP response from :func:`get()`.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     prev_summary = None
@@ -141,7 +141,7 @@ def _assert_success(summary: Dict) -> Optional[Dict]:
         request.
     :type summary: Dict
     :return: The summary.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     if "error_msg" in summary and summary["error_msg"]:
@@ -153,25 +153,61 @@ def _assert_success(summary: Dict) -> Optional[Dict]:
     return summary
 
 
-def _read_response_summary(resp: http.client.HTTPResponse) -> Optional[Dict]:
+# as used by register_summary_string_hook
+API_SUMMARY_HOOK_CALLBACK = None
+
+
+def register_summary_hook(function: Callable) -> None:
+    """
+    This function allows you to provide a function that will be called with
+    every summary object that is returned from any API call to the /api
+    endpoint, that being the raw response body before JSON parsing. This can be
+    used, for example, to instrument a test suite with a function that writes
+    every summary object to disk and then creating a dataset of Watchful summary
+    objects for further analysis. Most SDK users probably won't be reaching for
+    this function every day, but if you find a clever use for it, let us know!
+
+    :param function: Your function to be called with every summary string
+    :type function: Callable
+    """
+    global API_SUMMARY_HOOK_CALLBACK
+    API_SUMMARY_HOOK_CALLBACK = function
+
+
+def _read_response(
+    resp: http.client.HTTPResponse, resp_is_summary: bool = False
+) -> Optional[Dict]:
     """
     This function raises an exception if ``resp.status`` is not 200, otherwise
     it returns ``ret``.
 
+    If ``resp_is_summary`` then we will also run the
+    ``API_SUMMARY_HOOK_CALLBACK`` (any hook that was provided), so this is only
+    appropriate for endpoints that return a summary, that is, the /api endpoint
+    rather than some other JSON object like /config, /remote, etc.
+
     :param resp: The HTTP response from a connection request.
     :type resp: http.client.HTTPResponse
+    :param resp_is_summary: Boolean indicating if the response is known to be a
+        summary object.
+    :type resp_is_summary: bool, optional
     :return: The dictionary of ``resp``.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
+    # the assertion is here because that's what our API endpoints always return
+    assert 200 == resp.status
+    json_str = resp.read()
 
-    assert 200 == int(resp.status)
+    if resp_is_summary and API_SUMMARY_HOOK_CALLBACK:
+        API_SUMMARY_HOOK_CALLBACK(json_str)
+
+    ret = json.loads(json_str)
 
-    ret = json.loads(resp.read())
     # One idea:
     # if ret["error_msg"]:
     #     raise Exception(ret["error_msg"])
     if "error_msg" in ret and ret["error_msg"]:
-        print(ret["error_msg"])
+        print(f'{ret["error_verb"]}: {ret["error_msg"]}')
 
     return ret
 
@@ -186,7 +222,7 @@ def api(verb: str, **args: Dict) -> Optional[Dict]:
     :param args: Optional parameters to support the API for ``verb``.
     :type args: Dict
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     action = args  # already a dictionary
@@ -202,14 +238,14 @@ def api_send_action(action: Dict) -> Optional[Dict]:
     :param action: The ``verb`` for the API with optional parameters.
     :type action: Dict
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     conn = _get_conn()
     conn.request(
         "POST", "/api", json.dumps(action), {"Content-Type": "application/json"}
     )
-    return _read_response_summary(conn.getresponse())
+    return _read_response(conn.getresponse(), resp_is_summary=True)
 
 
 def ephemeral(port: str = "9002") -> None:
@@ -222,7 +258,8 @@ def ephemeral(port: str = "9002") -> None:
     """
 
     _ = spawn_cmd(
-        f"watchful -p {port} --no-persistence >watchful_ephemeral_output.txt 2>&1"
+        f"watchful -p {port} --no-persistence "
+        f">watchful_ephemeral_output.txt 2>&1"
     )
     await_port_opening(int(port))
     external(port=port)
@@ -258,7 +295,7 @@ def list_projects() -> Dict:
     return json.loads(r.read())
 
 
-def open_project(id_: str) -> bytes:
+def open_project(id_: str) -> str:
     """
     This function opens a project via its project id, which is the path to its
     hints file.
@@ -266,7 +303,7 @@ def open_project(id_: str) -> bytes:
     :param id_: The project id.
     :type id_: str
     :return: The read HTTP response.
-    :rtype: bytes
+    :rtype: str
     """
 
     c = _get_conn()
@@ -278,21 +315,44 @@ def open_project(id_: str) -> bytes:
     )
     r = c.getresponse()
 
-    ret = r.read()
+    ret = r.read().decode("utf-8")
     await_plabels()
 
     return ret
 
 
-def create_project() -> bytes:
+def create_project(title_: Optional[str] = None) -> Union[str, Optional[Dict]]:
     """
-    This function creates a new project.
+    This function creates a new project. Additionally, if title is supplied, a
+    title is given to the newly created project.
 
-    :return: The read HTTP response.
-    :rtype: bytes
+    :param title: The title for the project.
+    :type title: str, optional
+    :return: If a title is supplied and ``open_project("new")`` is successful,
+        the dictionary of the HTTP response from the connection request from
+        ``title(title)``; otherwise the read HTTP response from
+        ``open_project("new")``.
+    :rtype: Union[str, Optional[Dict]]
     """
 
-    return open_project("new")
+    ret = open_project("new")
+
+    if title_ is not None and ret == '"OK"\n':
+        return title(title_)
+    return ret
+
+
+def title(title_: str) -> Optional[Dict]:
+    """
+    This function gives a title to a newly created project.
+
+    :param title_: The title for the project.
+    :type title_: str
+    :return: The dictionary of the HTTP response from the connection request.
+    :rtype: Dict, optional
+    """
+
+    return api("title", title=title_)
 
 
 def get_project_id(summary: Dict) -> str:
@@ -435,24 +495,105 @@ def records(csv_: str) -> Optional[Dict]:
     :param csv_: The csv dataset.
     :type csv_: str
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("records", data=csv_, content_type="text/csv")
 
 
-def col_flag(flag: str, columns: List[bool]) -> Optional[Dict]:
+def _col_flag(
+    columns: List[bool],
+    flag: Literal["inferenceable"] = "inferenceable",
+) -> Optional[Dict]:
     """
-    This function sets a flag for all the columns in the list.
+    This function sets a flag for each of the columns of the dataset in the
+    currently active project.
 
-    :param flag: "inferenceable" is currently the only supported flag
-    :param columns: list of true/false values, specifying whether the flag should be set for the column
+    :param columns: A list of true/false values, specifying whether the flag
+        should be set for each column.
+    :type columns: List
+    :param flag: The flag to be set; "inferenceable" is currently the only
+        supported flag.
+    :type flag: str, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
+
     return api("col_flag", flag=flag, columns=columns)
 
 
+def set_col_flag(
+    columns: Optional[List[str]] = None,
+    flag: Literal["inferenceable"] = "inferenceable",
+    pos_sense: bool = True,
+) -> Optional[Dict]:
+    """
+    This function sets a flag for each of the columns of the dataset in the
+    currently active project. As a default, given columns will be set to
+    ``True`` and all other columns will be set to ``False``. A helper to
+    indicate all available columns as given is to omit ``columns``.
+
+    :param flag: The flag to be set; "inferenceable" is currently the only
+        supported flag.
+    :type flag: str, optional
+    :param columns: A list of column names specifying whether the flag should
+        be set.
+    :type columns: List, optional
+    :param pos_sense: A boolean specifying whether the setting of the flag
+        is in positive or negative sense.
+    :type pos_sense: bool, optional
+    :return: The dictionary of the HTTP response from the connection request.
+    :rtype: Dict, optional
+    """
+
+    col_names = get()["field_names"]
+    n_col_names = len(col_names)
+
+    if columns is None:
+        col_bools = [pos_sense] * n_col_names
+    else:
+        n_columns = len(columns)
+
+        assert (
+            len(set(columns)) == n_columns
+        ), f"At least one of the given columns in {columns} is duplicate!"
+        assert len(set(columns) - set(col_names)) == 0, (
+            f"At least one of the given columns in {columns} is not available "
+            f"in the dataset columns {col_names}!"
+        )
+
+        def __f(x):
+            i = x in columns
+            return i if pos_sense else not i
+
+        col_bools = list(map(__f, col_names))
+
+    return _col_flag(col_bools, flag)
+
+
+def ignore_col_flag(
+    columns: Optional[List[str]] = None,
+    flag: Literal["inferenceable"] = "inferenceable",
+) -> Optional[Dict]:
+    """
+    This function sets a flag for each of the columns of the dataset in the
+    currently active project. Given columns will be set to ``False`` and all
+    other columns will be set to ``True``. A helper to indicate all available
+    columns as given is to omit ``columns``.
+
+    :param flag: The flag to be set; "inferenceable" is currently the only
+        supported flag.
+    :type flag: str, optional
+    :param columns: A list of column names specifying whether the flag should
+        be set to ``False``.
+    :type columns: List, optional
+    :return: The dictionary of the HTTP response from the connection request.
+    :rtype: Dict, optional
+    """
+
+    return set_col_flag(columns, flag, False)
+
+
 def class_(class__: str) -> Optional[Dict]:
     """
     This function creates a class.
@@ -460,7 +601,7 @@ def class_(class__: str) -> Optional[Dict]:
     :param class__: The class.
     :type class__: str
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("class", name=class__)
@@ -476,7 +617,7 @@ def create_class(class__: str, class_type: str = "ftc") -> Optional[Dict]:
         to "ftc".
     :type class_type: str, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("class", name=class__, class_type=class_type)
@@ -492,7 +633,7 @@ def query_async(q: str, page: int = 0) -> Optional[Dict]:
     :param page: The page, defaults to 0.
     :type page: int, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("query", query=q, page=page)
@@ -507,7 +648,7 @@ def query(q: str, page: int = 0) -> Optional[Dict]:
     :param page: The page, defaults to 0.
     :type page: int, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     query_async(q, page)
@@ -553,7 +694,7 @@ def base_rate(class__: str, rate: int) -> Optional[Dict]:
     :param rate: The base rate for the class.
     :type rate: int
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("base_rate", label=class__, rate=rate)
@@ -564,7 +705,7 @@ def await_plabels() -> Optional[Dict]:
     This function gets the updated HTTP response.
 
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return await_summary(lambda s: s["status"] == "current")
@@ -582,7 +723,7 @@ def hinter_async(class__: str, query_: str, weight: int) -> Optional[Dict]:
     :param weight: The weight for the hinter.
     :type weight: int
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("hinter", label=class__, weight=weight, query=query_)
@@ -599,7 +740,7 @@ def hinter(class__: str, query_: str, weight: int) -> Optional[Dict]:
     :param weight: The weight for the hinter.
     :type weight: int
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     _assert_success(hinter_async(class__, query_, weight))
@@ -613,7 +754,7 @@ def delete(id_: int) -> Optional[Dict]:
     :param id_: The hinter id to delete.
     :type id_: int
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("delete", id=id_)
@@ -626,7 +767,7 @@ def delete_class(class__: str) -> Optional[Dict]:
     :param class__: The class to delete.
     :type class__: str
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("delete", class_name=class__)
@@ -641,25 +782,12 @@ def get() -> Optional[Dict]:
     more.
 
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("nop")
 
 
-def title(title_: str) -> Optional[Dict]:
-    """
-    This function gives a title to a newly created project.
-
-    :param title_: The title for the project.
-    :type title_: str
-    :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
-    """
-
-    return api("title", title=title_)
-
-
 def external_hinter(class__: str, name: str, weight: int) -> Optional[Dict]:
     """
     This function creates an external hinter.
@@ -671,7 +799,7 @@ def external_hinter(class__: str, name: str, weight: int) -> Optional[Dict]:
     :param weight: The weight for the hinter.
     :type weight: int
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     _assert_success(
@@ -701,7 +829,7 @@ def upload_attributes(
     :param attributes_filepath: The attributes filepath.
     :type attributes_filepath: str
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
     conn = _get_conn()
     with open(attributes_filepath, encoding="utf-8") as attributes_file:
@@ -713,7 +841,7 @@ def upload_attributes(
         )
     resp = conn.getresponse()
     assert resp.status == 200, f"not OK HTTP status. Was: {resp.status}"
-    return _assert_success(_read_response_summary(resp))
+    return _assert_success(_read_response(resp))
 
 
 def load_attributes(
@@ -729,7 +857,7 @@ def load_attributes(
     :param attributes_filename: The attributes filename.
     :type attributes_filename: str
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("attributes", id=dataset_id, file=attributes_filename)
@@ -742,7 +870,7 @@ def _dump(offset: int) -> Optional[Dict]:
     :param offset: The offset chunk.
     :type offset: int
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("dump", offset=offset)
@@ -796,7 +924,7 @@ def hint(name: str, offset: int, values: List[bool]) -> Optional[Dict]:
     :param values: The hints.
     :type values: List[bool]
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
 
     TODO: Come up with a better streaming Python API here.
     """
@@ -813,7 +941,7 @@ def apply_hints(name: str) -> Optional[Dict]:
     :param name: The hinter name.
     :type name: str
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     _assert_success(api("apply_hints", name=name))
@@ -830,7 +958,7 @@ def hint_all(name: str, values: List[bool]) -> Optional[Dict]:
     :param values: The hints.
     :type values: List[bool]
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     hint(name, 0, values)
@@ -871,7 +999,7 @@ def export_stream(
         + urllib.parse.quote_plus(mode),
     )
     resp = conn.getresponse()
-    assert 200 == int(resp.status), f"Request failed with status {resp.status}."
+    assert 200 == resp.status, f"Request failed with status {resp.status}."
     return resp
 
 
@@ -919,7 +1047,7 @@ def export_async() -> Optional[Dict]:
     response is likely not updated yet.
 
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("export", query="", content_type="text/csv")
@@ -930,7 +1058,7 @@ def export() -> Optional[Dict]:
     This function exports the dataset and returns an updated HTTP response.
 
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     n_exports = len(get()["exports"])
@@ -947,21 +1075,102 @@ def export_preview(mode: str = "ftc") -> Optional[Dict]:
         "ner", defaults to "ftc".
     :type mode: str, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return api("export_preview", mode=mode)
 
 
+def export_project() -> http.client.HTTPResponse:
+    """
+    This function returns a consolidated version (a single *.hints file) of the
+    currently open project. Unlike other GET endpoints that return a summary
+    object, this function returns a streamed file, the *.hints project file.
+
+    :return:  The HTTP response from the connection request.
+    :rtype: http.client.HTTPResponse
+    """
+
+    conn = _get_conn()
+    conn.request("GET", "/export_project")
+    resp = conn.getresponse()
+    assert 200 == resp.status, f"Request failed with status {resp.status}."
+    return resp
+
+
+def is_utf8(
+    csv_bytes: bytes = None,
+    filepath: str = None,
+    threshold: float = 0.5,
+    is_fast: bool = True,
+) -> bool:
+    """
+    This function attempts to detect if the encoding of the given bytes or the
+    content of the given filepath is utf-8. It returns `True` if the detected
+    encoding is utf-8 and has a confidence of the given threshold or more,
+    otherwise `False`. This function may need some tweaking for a very large
+    dataset, but should work with the ``is_fast`` argument set to `True` by
+    default.
+
+    :param csv_bytes: The csv dataset bytes.
+    :type csv_bytes: bytes
+    :param filepath: The path of the csv dataset file.
+    :type filepath: str
+    :param threshold: The minimum confidence required to accept the detected
+        encoding.
+    :type threshold: float, optional
+    :param is_fast: Whether to use fast encoding detection with a lower
+        accuracy, or not.
+    :type is_fast: bool, optional
+    :return: `True` if the detected encoding is utf-8 and has a confidence of
+        the given threshold or more, otherwise `False`.
+    :rtype: bool
+    """
+    if csv_bytes is None and filepath is None:
+        raise Exception(
+            "Both filepath and csv_bytes are not specified. "
+            "One of them needs to be specified."
+        )
+    if csv_bytes and filepath:
+        raise Exception(
+            "Both filepath and csv_bytes are specified. "
+            "Only one of them needs to be specified."
+        )
+
+    if is_fast:
+        import cchardet as chardet
+    else:
+        import chardet
+
+    if csv_bytes:
+        res = chardet.detect(csv_bytes)
+    else:
+        if os.path.isfile(filepath):
+            with open(filepath, "rb") as f:
+                res = chardet.detect(f.read())
+        else:
+            raise Exception(
+                f"There is no file at the given file path {filepath}!"
+            )
+
+    if res["encoding"].lower() == "utf-8" and res["confidence"] >= threshold:
+        return True
+    return False
+
+
 def create_dataset(
     csv_bytes: bytes,
     columns: List[str],
     filename: str = "none",
     has_header: bool = True,
+    threshold_detect: float = 0.5,
+    is_fast_detect: bool = True,
+    force_load: bool = False,
 ) -> str:
     """
     This function loads the specified columns of a csv dataset and returns the
-    dataset id.
+    dataset id if its encoding is detected to be utf-8 or if dataset loading is
+    forced.
 
     :param csv_bytes: The csv dataset bytes.
     :type csv_bytes: bytes
@@ -972,34 +1181,56 @@ def create_dataset(
     :param has_header: The boolean indicating if the csv dataset has a header,
         defaults to True.
     :type has_header: bool, optional
-    :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :param threshold_detect: The minimum confidence required to accept the
+        detected encoding.
+    :type threshold_detect: float, optional
+    :param is_fast_detect: Whether to use fast encoding detection with a lower
+        accuracy, or not.
+    :type is_fast_detect: bool, optional
+    :param force_load: The boolean indicating if the csv dataset will be loaded
+        even when its encoding is detected to be non-utf-8, defaults to False.
+        This is useful in rare cases where the csv dataset is detected to be
+        non-utf-8 encoded and the user is sure about the csv dataset being utf-8
+        encoded.
+    :type force_load: bool, optional
+    :return: The dataset id.
+    :rtype: str
 
     TODO: Add error handling.
     """
 
-    id_ = str(uuid4())
-    conn = _get_conn()
-    conn.request(
-        "POST",
-        "/api/_stream/" + id_ + "/0/true",
-        csv_bytes,
-        {"Content-Type": "text/csv"},
+    is_csv_bytes_utf8 = is_utf8(
+        csv_bytes, None, threshold_detect, is_fast_detect
     )
-    _ = _read_response_summary(conn.getresponse())
 
-    params = json.dumps({"filename": filename, "has_header": has_header})
-    conn.request(
-        "POST",
-        "/api/_stream/" + id_,
-        params,
-        {"Content-Type": "application/json"},
+    if is_csv_bytes_utf8 or force_load:
+        id_ = str(uuid4())
+        conn = _get_conn()
+        conn.request(
+            "POST",
+            "/api/_stream/" + id_ + "/0/true",
+            csv_bytes,
+            {"Content-Type": "text/csv"},
+        )
+        _ = _read_response(conn.getresponse())
+
+        params = json.dumps({"filename": filename, "has_header": has_header})
+        conn.request(
+            "POST",
+            "/api/_stream/" + id_,
+            params,
+            {"Content-Type": "application/json"},
+        )
+        dataset_id = _read_response(conn.getresponse())["id"]
+
+        api("dataset_add", id=dataset_id, columns=columns)
+
+        return dataset_id
+
+    raise Exception(
+        "Dataset is not loaded as the encoding of the csv dataset is not "
+        "detected to be utf-8 and dataset loading is not forced."
     )
-    dataset_id = _read_response_summary(conn.getresponse())["id"]
-
-    api("dataset_add", id=dataset_id, columns=columns)
-
-    return dataset_id
 
 
 def label_single(row: List[str]) -> List[str]:
@@ -1044,7 +1275,7 @@ def config_set(key: str, value: str) -> Optional[Dict]:
     :param value: The parameter value.
     :type value: str
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     conn = _get_conn()
@@ -1053,7 +1284,7 @@ def config_set(key: str, value: str) -> Optional[Dict]:
         "POST", "/config", params, {"Content-Type": "application/json"}
     )
 
-    return _read_response_summary(conn.getresponse())
+    return _read_response(conn.getresponse())
 
 
 def config() -> Optional[Dict]:
@@ -1062,13 +1293,28 @@ def config() -> Optional[Dict]:
     ``remote``, ``username``, ``role`` and ``authorization`` and their values.
 
     :return: A dictionary of key value pairs
-    :rtype: dict
+    :rtype: Dict, optional
     """
 
     conn = _get_conn()
     conn.request("GET", "/config", None, {"Content-Type": "application/json"})
 
-    return _read_response_summary(conn.getresponse())
+    return _read_response(conn.getresponse())
+
+
+def set_hub_url(url: str) -> Optional[Dict]:
+    """
+    This function sets the Watchful hub URL of the Watchful client. The Watchful
+    hub URL should not change after data has been fetched or published to a hub.
+
+    :param url: The Watchful hub URL.
+    :type url: str
+    :return: The dictionary of the HTTP response from the connection request.
+    :rtype: Dict, optional
+    """
+    conn = _get_conn()
+    conn.request("POST", "/set_hub_url", url, {"Content-Type": "text/plain"})
+    return _read_response(conn.getresponse())
 
 
 def print_candidates(summary: Optional[Dict] = None) -> None:
@@ -1117,14 +1363,14 @@ def exit_backend() -> None:
     """
     This function exits the backend. Note that the API call will usually fail
     because the backend exits before returning a HTTP response so we suppress
-    the error. This is an old API function wrapper and will not be applicable to
-    the dockerized Watchful application instances.
+    the error. This is useful locally, for tests, and during development, but
+    not in dockerized Watchful application instances.
     """
 
     try:
         api("exit")
-    # We allowing passing this error as it is known.
-    except http.client.RemoteDisconnected:
+    # see docstring above
+    except (http.client.RemoteDisconnected, http.client.IncompleteRead):
         pass
 
 
@@ -1143,7 +1389,7 @@ def hub_api(verb: str, token: str, **args: Dict) -> Optional[Dict]:
     :param args: Optional parameters to support the hub API for ``verb``.
     :type args: Dict
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     headers = {"Content-Type": "application/json"}
@@ -1152,7 +1398,7 @@ def hub_api(verb: str, token: str, **args: Dict) -> Optional[Dict]:
     action["verb"] = verb
     conn = _get_conn()
     conn.request("POST", "/remote", json.dumps(action), headers)
-    return _assert_success(_read_response_summary(conn.getresponse()))
+    return _assert_success(_read_response(conn.getresponse()))
 
 
 def login(email: str, password: str) -> Optional[Dict]:
@@ -1164,7 +1410,7 @@ def login(email: str, password: str) -> Optional[Dict]:
     :param password: The user's password.
     :type password: str
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     headers = {"Content-Type": "application/json"}
@@ -1174,7 +1420,7 @@ def login(email: str, password: str) -> Optional[Dict]:
     headers.update({"Authorization": f"Basic {credentials}"})
     conn = _get_conn()
     conn.request("POST", "/remote", json.dumps({"verb": "login"}), headers)
-    return _assert_success(_read_response_summary(conn.getresponse()))
+    return _assert_success(_read_response(conn.getresponse()))
 
 
 def publish(token: Optional[str] = None) -> Optional[Dict]:
@@ -1184,7 +1430,7 @@ def publish(token: Optional[str] = None) -> Optional[Dict]:
     :param token: The user's auth token, defaults to None.
     :type token: str, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return hub_api("publish", token)
@@ -1197,7 +1443,7 @@ def fetch(token: Optional[str] = None) -> Optional[Dict]:
     :param token: The user's auth token, defaults to None.
     :type token: str, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return hub_api("fetch", token)
@@ -1210,7 +1456,7 @@ def pull(token: Optional[str] = None) -> Optional[Dict]:
     :param token: The user's auth token, defaults to None.
     :type token: str, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return hub_api("pull", token)
@@ -1223,7 +1469,7 @@ def push(token: Optional[str] = None) -> Optional[Dict]:
     :param token: The user's auth token, defaults to None.
     :type token: str, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return hub_api("push", token)
@@ -1236,7 +1482,7 @@ def peek(token: Optional[str] = None) -> Optional[Dict]:
     :param token: The user's auth token, defaults to None.
     :type token: str, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return hub_api("peek", token)
@@ -1249,7 +1495,7 @@ def whoami(token: Optional[str] = None) -> Optional[Dict]:
     :param token: The user's auth token, defaults to None.
     :type token: str, optional
     :return: The dictionary of the HTTP response from the connection request.
-    :rtype: Dict
+    :rtype: Dict, optional
     """
 
     return hub_api("whoami", token)
