@@ -12,7 +12,6 @@ import pathlib
 import shutil
 import sys
 from typing import List, Type
-from uuid import uuid4
 from watchful import client, attributes
 from watchful.enricher import Enricher
 
@@ -172,7 +171,7 @@ def main(args: List[str] = None, custom_enricher: Enricher = None) -> None:
             user_home_path = os.path.expanduser("~")
             working_dir = os.path.join(user_home_path, "watchful", "working")
             os.makedirs(working_dir, exist_ok=True)
-            args.in_file = os.path.join(working_dir, str(uuid4()))
+            args.in_file = os.path.join(working_dir, dataset_id)
             client.export_dataset_to_path(args.in_file, summary["field_names"])
         else:
             print(
@@ -397,8 +396,9 @@ def main(args: List[str] = None, custom_enricher: Enricher = None) -> None:
         # Example usage:
         # curl -iX POST http://localhost:9001/api \
         #   --header "Content-Type: application/json" \
-        #   --data '{"verb":"attributes","id":"9570b0b5-4a58-445f-9b51-b434caca2650",
-        #         "filepath":"/path/to/attributes_file.attrs"}'
+        #   --data "{\"verb\":\"attributes\",\
+        # \"id\":\"9570b0b5-4a58-445f-9b51-b434caca2650\",\
+        # \"filepath\":\"/path/to/attributes_file.attrs\"}"
         # Arguments:
         #     id: dataset id
         #     filepath: attributes filepath
